# coding=utf-8

from typing import Optional, Any, Union, AnyStr


class TioResponse:
    _code = 0
    _data = None
    _result = None
    _error = None

    def __init__(self, code, data=None, error=None):
        # type: (Union[int, AnyStr], Optional[Any], Optional[Any]) -> None
        self._code = code
        self._data = data
        if data is None:
            self._splitdata = [None, error]
        else:
            self._splitdata = self._data.split(self._data[:16])

        if not self._splitdata[1] or self._splitdata[1] == b'':
            self._error = b''.join(self._splitdata[2:])
            self._result = None
        else:
            self._error = None
            self._result = self._splitdata[1]

    @property
    def code(self):
        # type: () -> Union[int, AnyStr]
        return self._code.decode('utf-8')

    @property
    def result(self):
        # type: () -> Optional[AnyStr]
        if self._result:
            return self._result.decode('utf-8')
        else:
            return None

    @property
    def error(self):
        # type: () -> Optional[AnyStr]
        if self._error:
            return self._error.decode('utf-8')
        else:
            return None

    @property
    def raw(self):
        # type: () -> Any
        return self._data

    def get_code(self):
        # type: () -> Union[int, AnyStr]
        return self.code

    def get_result(self):
        # type: () -> Optional[AnyStr]
        return self.result

    def get_error(self):
        # type: () -> Optional[AnyStr]
        return self.error



# coding=utf-8
import gzip
import io
import json
import platform
from typing import AnyStr, Union

from ._TioRequest import TioRequest
from ._TioResponse import TioResponse

# Version specific import handling.
if platform.python_version() <= '3.0':
    # noinspection PyCompatibility,PyUnresolvedReferences
    from urllib2 import urlopen
    # noinspection PyCompatibility,PyUnresolvedReferences
    from urllib2 import HTTPError, URLError
else:
    # noinspection PyCompatibility
    from urllib.request import urlopen
    # noinspection PyCompatibility
    from urllib.error import HTTPError, URLError


class Tio:
    backend = "cgi-bin/run/api/"
    json = "languages.json"

    def __init__(self, url="https://tio.run"):
        # type: (AnyStr) -> None
        self.backend = url + '/' + self.backend
        self.json = url + '/' + self.json

    @staticmethod
    def read_in_chunks(stream_object, chunk_size=1024):
        """Lazy function (generator) to read a file piece by piece.
        Default chunk size: 1k."""
        while True:
            data = stream_object.read(chunk_size)
            if not data:
                break
            yield data

    @staticmethod
    def new_request(lang, code):
        # type: (AnyStr, Union[AnyStr, bytes]) -> TioRequest
        return TioRequest(lang=lang, code=code)

    def query_languages(self):
        # type: () -> set
        try:
            response = urlopen(self.json)
            rawdata = json.loads(response.read().decode('utf-8'))
            return set(rawdata.keys())
        except (HTTPError, URLError):
            return set()
        except Exception:
            return set()

    def send(self, fmt):
        # type: (TioRequest) -> TioResponse
        return self.send_bytes(fmt.as_deflated_bytes())

    def send_bytes(self, message):
        # type: (bytes) -> TioResponse
        req = urlopen(self.backend, data=message)
        reqcode = req.getcode()
        if req.code == 200:
            if platform.python_version() >= '3.0':
                content_type = req.info().get_content_type()
            else:
                content_type = req.info()['content-type']

            if content_type == 'application/octet-stream':
                buf = io.BytesIO(req.read())
                gzip_f = gzip.GzipFile(fileobj=buf)
                fulldata = gzip_f.read()
            else:
                fulldata = req.read()

            return TioResponse(reqcode, fulldata, None)
        else:
            return TioResponse(reqcode, None, None)



# coding=utf-8

from ._Tio import Tio
from ._TioFile import TioFile
from ._TioRequest import TioRequest
from ._TioResponse import TioResponse
from ._TioResult import TioResult
from ._TioVariable import TioVariable

__title__ = 'PyTIO'
__author__ = 'Thomas Ward'
__version__ = '0.1.0'
__copyright__ = '2017 Thomas Ward'
__license__ = 'AGPLv3+'
__all__ = ('Tio', 'TioFile', 'TioRequest', 'TioResponse', 'TioResult', 'TioVariable')



# coding=utf-8

import platform
import zlib
from typing import List, AnyStr, Union
from ._TioFile import TioFile
from ._TioVariable import TioVariable


class TioRequest:

    def __init__(self, lang=None, code=None):
        # type: (AnyStr, Union[AnyStr, bytes]) -> None
        self._files = []
        self._variables = []
        self._bytes = bytes()

        if lang:
            self.set_lang(lang)

        if code:
            self.add_file_bytes('.code.tio', code)

    def add_file(self, file):
        # type: (TioFile) -> None
        if file in self._files:
            self._files.remove(file)
        self._files.append(file)

    def add_file_bytes(self, name, content):
        # type: (AnyStr, bytes) -> None
        self._files.append(TioFile(name, content))

    def add_variable(self, variable):
        # type: (TioVariable) -> None
        self._variables.append(variable)

    def add_variable_string(self, name, value):
        # type: (AnyStr, Union[List[AnyStr], AnyStr]) -> None
        self._variables.append(TioVariable(name, value))

    def set_lang(self, lang):
        # type: (AnyStr) -> None
        self.add_variable_string('lang', lang)

    def set_code(self, code):
        # type: (AnyStr) -> None
        self.add_file_bytes('.code.tio', code)

    def set_input(self, input_data):
        # type: (AnyStr) -> None
        self.add_file_bytes('.input.tio', input_data.encode('utf-8'))

    def set_compiler_flags(self, flags):
        # type: (AnyStr) -> None
        self.add_variable_string('TIO_CFLAGS', flags)

    def set_commandline_flags(self, flags):
        # type: (AnyStr) -> None
        self.add_variable_string('TIO_OPTIONS', flags)

    def set_arguments(self, args):
        # type: (AnyStr) -> None
        self.add_variable_string('args', args)

    def write_variable(self, name, content):
        # type: (AnyStr, AnyStr) -> None
        if content:
            if platform.python_version() >= '3.0':
                self._bytes += bytes("V" + name + '\x00' + str(len(content.split(' '))) + '\x00', 'utf-8')
                self._bytes += bytes(content + '\x00', 'utf-8')
            else:
                self._bytes += bytes("V" + name + '\x00' + str(len(content.split(' '))) + '\x00')
                self._bytes += bytes(content + '\x00')

    def write_file(self, name, contents):
        # type: (AnyStr, AnyStr) -> None
        # noinspection PyUnresolvedReferences
        if platform.python_version() < '3.0' and isinstance(contents, str):
            length = len(contents)
        elif isinstance(contents, str):
            length = len(contents.encode('utf-8'))
        elif isinstance(contents, (bytes, bytearray)):
            length = len(contents)
        else:
            raise ValueError("Can only pass UTF-8 strings or bytes at this time.")
        if platform.python_version() >= '3.0':
            self._bytes += bytes("F" + name + '\x00' + str(length) + '\x00', 'utf-8')
            self._bytes += bytes(contents + '\x00', 'utf-8')
        else:
            self._bytes += bytes("F" + name + '\x00' + str(length) + '\x00')
            self._bytes += bytes(contents + '\x00')

    def as_bytes(self):
        # type: () -> bytes
        try:
            for var in self._variables:
                if hasattr(var, 'name') and hasattr(var, 'content'):
                    self.write_variable(var.name, var.content)

            for file in self._files:
                if hasattr(file, 'name') and hasattr(file, 'content'):
                    self.write_file(file.name, file.content)

            self._bytes += b'R'

        except IOError:
            raise RuntimeError("IOError generated during bytes conversion.")

        return self._bytes

    def as_deflated_bytes(self):
        # type: () -> bytes
        return zlib.compress(self.as_bytes(), 9)[2:-4]



# coding=utf-8

from typing import Any, AnyStr, List


class TioResult:
    _pieces = []

    def __init__(self, pieces):
        # type: (List[AnyStr]) -> None
        self._pieces = pieces

        raise NotImplementedError

    @property
    def pieces(self):
        # type: () -> List
        return self._pieces

    def has(self, field):
        # type: (AnyStr) -> bool
        try:
            if field.lower() == "output":
                return len(self._pieces) > 0
            elif field.lower() == "debug":
                return len(self._pieces) > 1
            else:
                return False
        except IndexError:
            return False

    def get(self, field):
        # type: (AnyStr) -> Any
        if self.has('output') and field.lower() == "output":
            return self._pieces[0]
        elif self.has('debug') and field.lower() == "debug":
            return self._pieces[1]



# coding=utf-8
from setuptools import find_packages, setup

# noinspection PyProtectedMember
from pytio import __version__ as version

setup(
    name='pytio',
    version=version,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'typing',
    ],
    author='Thomas Ward',
    author_email='teward@dark-net.io',
    description="A library to interact with TIO.run and execute code, as well as return the results for the execution.",
    long_description="A library to interact with TIO.run and execute code, as well as return the results for the "
                     "execution. This is basically a Python version of https://github.com/SocraticPhoenix/TioJ but "
                     "using Pythonic approaches and software rather than the complex Java solutions.",
    license='AGPLv3+',
    url='https://github.com/teward/pytio',
    download_url='https://pypi.python.org/pypi/pytio',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Communications :: Email',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='tryitonline tio',
    platforms='any',
    test_suite='test_tio',
)



# coding=utf-8

from typing import AnyStr, List, Union


class TioVariable:
    _name = str()
    _content = []

    def __init__(self, name, content):
        # type: (AnyStr, Union[List[AnyStr], AnyStr]) -> None
        self._name = name
        self._content = content

    @property
    def name(self):
        # type: () -> AnyStr
        return self._name

    @property
    def content(self):
        # type: () -> List
        return self._content

    def get_name(self):
        # type: () -> AnyStr
        return self.name

    def get_content(self):
        # type: () -> List
        return self.content
